/*
 *  Authors:
 *    Per Brand, Konstantin Popov
 * 
 *  Contributors:
 * 
 *  Copyright:
 *    Per Brand, Konstantin Popov 1998
 * 
 *  Last change:
 *    $Date: 1999/09/10 13:23:20 $ by $Author: andreas $
 *    $Revision: 1.23 $
 * 
 *  This file is part of Mozart, an implementation 
 *  of Oz 3:
 *     http://www.mozart-oz.org
 * 
 *  See the file "LICENSE" or
 *     http://www.mozart-oz.org/LICENSE.html
 *  for information on usage and redistribution 
 *  of this file, and for a DISCLAIMER OF ALL 
 *  WARRANTIES.
 *
 */

#ifndef __DPINTERFACE_HH
#define __DPINTERFACE_HH

#ifdef INTERFACE
#pragma interface
#endif

#include "base.hh"
#include "pickle.hh"

#define SIZEOFPORTPROXY (4*sizeof(void*))

//
extern Bool (*isPerdioInitialized)();
 
// 
extern OZ_Return (*portSend)(Tertiary *p, TaggedRef msg);
extern OZ_Return (*cellDoExchange)(Tertiary*,TaggedRef,TaggedRef);
extern OZ_Return (*cellDoAccess)(Tertiary*,TaggedRef);
extern OZ_Return (*cellAtAccess)(Tertiary*,TaggedRef,TaggedRef);
extern OZ_Return (*cellAtExchange)(Tertiary*,TaggedRef,TaggedRef);
extern OZ_Return (*cellAssignExchange)(Tertiary*,TaggedRef,TaggedRef);
extern OZ_Return (*objectExchange) (Tertiary*,TaggedRef,TaggedRef,TaggedRef);

// lock/unlock (interface) methods/their usage may be optimized
// further, e.g. inline cases when distributed locks are currently
// local;
extern void (*lockLockProxy)(Tertiary *t, Thread *thr);
extern LockRet (*lockLockManagerOutline)(LockManagerEmul *lfu, Thread *thr);
extern void (*unlockLockManagerOutline)(LockManagerEmul *lfu, Thread *thr);
extern LockRet (*lockLockFrameOutline)(LockFrameEmul *lfu, Thread *thr);
extern void (*unlockLockFrameOutline)(LockFrameEmul *lfu, Thread *thr);

//
// An invariant is that all the 'marshal*' procedures build a
// representation with a corresponding 'DIF' tag, e.g. 'marshalObject'
// puts 'DIF_OBJECT' into stream.  In the current implementation,
// however, there is a limitation: marshaling tertiaries/object
// variables could yield 'DIF_OWNER*'.  This means that those entries
// must either be all remembered or not remembered;
extern void (*marshalTertiary)(Tertiary *t, MarshalTag tag, MsgBuffer *bs);
#ifdef USE_FAST_UNMARSHALER
extern OZ_Term (*unmarshalTertiary)(MsgBuffer *bs, MarshalTag tag);
extern OZ_Term (*unmarshalOwner)(MsgBuffer *bs,MarshalTag mt);
extern OZ_Term (*unmarshalVar)(MsgBuffer*,Bool,Bool);
#else
extern OZ_Term (*unmarshalTertiaryRobust)(MsgBuffer *bs,MarshalTag tag,int *);
extern OZ_Term (*unmarshalOwnerRobust)(MsgBuffer *bs,MarshalTag mt,int *error);
extern OZ_Term (*unmarshalVarRobust)(MsgBuffer*,Bool,Bool,int*);
#endif
extern Bool (*marshalVariable)(TaggedRef*, MsgBuffer*);
extern Bool (*triggerVariable)(TaggedRef*);
extern void (*marshalObject)(ConstTerm *t, MsgBuffer *bs);
extern void (*marshalSPP)(TaggedRef term, MsgBuffer *bs,Bool trail);

//
extern void (*gCollectProxyRecurse)(Tertiary *t);
extern void (*gCollectManagerRecurse)(Tertiary *t);
extern ConstTerm* (*gCollectDistResource)(ConstTerm*);
extern void (*gCollectDistCellRecurse)(Tertiary *t);
extern void (*gCollectDistLockRecurse)(Tertiary *t);
extern void (*gCollectDistPortRecurse)(Tertiary *t);
//
//
//
extern void (*gCollectBorrowTableUnusedFrames)();
extern void (*gCollectFrameToProxy)();

//
extern void (*gCollectPerdioFinal)();
extern void (*gCollectPerdioRoots)();
extern void (*gCollectEntityInfo)(Tertiary*);

// exit hook;
extern void (*dpExit)();

// Debug stuff;
#ifdef DEBUG_CHECK
extern void (*maybeDebugBufferGet)(BYTE b);
extern void (*maybeDebugBufferPut)(BYTE b);
#endif

// distribution handlers
extern Bool (*distHandlerInstall)(unsigned short,unsigned short,
				       Thread*,TaggedRef, TaggedRef);
extern Bool (*distHandlerDeInstall)(unsigned short,unsigned short,
				       Thread*,TaggedRef, TaggedRef);
#endif // __DPINTERFACE_HH


